<?php

namespace EventPortal\Domain\Util\IOEngage;

use App\Jobs\ProcessAttendance;
use Carbon\Carbon;
use EventPortal\Domain\Delegate\Models\Delegate;
use EventPortal\Domain\Event\Models\Event;
use EventPortal\Domain\Event\Models\EventDelegate;
use EventPortal\Domain\Util\IOEngage\Exceptions\AuthenticationFailed;
use GuzzleHttp\Client;
use Illuminate\Support\Arr;
use Psr\Http\Message\ResponseInterface;

class IOEngage
{
    /**
     * @var string
     */
    protected $exception;

    /**
     * @var Client
     */
    protected $http;

    public function __construct()
    {
        $this->http = new Client();
    }

    public function updateDelegateEvent(Event $event, Delegate $delegate): void
    {
        $this->makeRequest(
            config('ioengage.endpoints.meeting'),
            MeetingUpdateFailed::class,
            'post',
            $this->transformModels($event, $delegate),
        );

        // 
        // collect($delegates)->each(function (Delegate $delegate) use ($event): void {
        //     ProcessAttendance::dispatch($event, $delegate);
        // }); 
    }

    public function authenticate(): self
    {
        if ($this->needsToAuthenticate() === false) {
            return $this;
        }

        $this->exception = AuthenticationFailed::class;

        $url = $this->buildUrlParams(
            config('ioengage.endpoints.token'),
            config('ioengage.authentication'),
        );

        $response = $this->http->post($url);

        $payload = $this->getContents($response, ['access_token', 'expires_in']);

        cache(['ioengage' => [
            'token' => $payload->access_token,
            'expires' => Carbon::now()->addSeconds($payload->expires_in),
        ]]);

        return $this;
    }

    public function makeRequest(string $uri, string $exception, string $method = 'get', ?array $body = null): ResponseInterface
    {
        $this->authenticate();

        $this->exception = $exception;

        ['token' => $token] = cache('ioengage');

        return $this->http->$method(
            $uri,
            [
                'headers' => [
                    'Authorization' => "Bearer {$token}",
                ], 'body' => $body,
            ]
        );
    }

    private function getContents(ResponseInterface $response, array $keys, int $status = 200): object
    {
        throw_if(
            $response->getStatusCode() !== $status
                || $this->payloadContains($response, $keys),
            new $this->exception
        );

        return json_decode($response->getBody());
    }

    private function payloadContains(ResponseInterface $response, array $keys): bool
    {
        $decode = json_decode($response->getBody()->getContents(), true);

        return Arr::has(array_keys($decode), $keys);
    }

    /**
     * Checks the current cache for the existence of credentials, then checks the expiry to determine whether the token is still valid.
     */
    private function needsToAuthenticate(): bool
    {
        $cache = cache('ioengage');
        if ($cache !== null
            && Arr::get($cache, 'token') !== null
            && Arr::get($cache, 'expires') > Carbon::now()
        ) {
            return false;
        }

        return true;
    }

    private function buildUrlParams(string $baseUri, array $args): string
    {
        return "{$baseUri}?" . implode('&', collect($args)->map(function ($value, $key): string {
            return "{$key}={$value}";
        })->toArray());
    }

    private function transformModels(Event $event, Delegate $delegate): array
    {
        $eventDelegate = EventDelegate::forEventAndDelegate($event, $delegate)->first();

        $dates = $event->readableDates;
        return [
            'CountryCode' => 'GB', // define the static ones in config and merge them I think
            'DelegateStatus' => $delegate->readableStatus,
            'EventEndDateTime' => $dates[1]->format('Y-m-d:T:z'),
            'EventStartDateTime' => $dates[0]->format('Y-m-d:T:z'),
            'EventUrl' => route('event.landing.index', ['event' => $event], true),
            'InvitationStatus' => $eventDelegate->status,
            'MeetID' => $event->id,
            'MeetingChannel' => 'Virtual', // required with approval, needs to be set /new type added to event
            'MeetingName' => $event->name,
            'MeetingStatus' => $event->status,
            'MileStone' => 'N/A', // mile stone? r u okay?
            'OneKeyId' => $delegate->identifier,
            'Source' => 'MEET',
            'MeetingType' => $event->requiresApproval ? 'Meeting with approval' : 'Meeting without approval', // see big list
            'SubType' => $event->type, // required smoetimes 0
            'SubTypeActivity' => $event->activity, // required sometimes >> required sometimes 0
        ];
    }
}
